/* gcc -o cl4others cl4others.c -lncp */

/************************************************
cl4others - Client Trust for Others v0.9.0
By: Matthew W Whited (matt@whitedonline.com)

helped by:
Allen Williams (datavsyoda@aol.com)
Petr Vandrovec (vandrove@vc.cvut.cz)

=================================================

This application allows for slient authentication
to Novell Bordermanager servers using the single
sign-on protocol called ClientTrust.  This was 
redeveloped for use with Linux and NCPFS. 

Please contact Matthew Whited at matt@whitedonline.com
if you find any bugs.  Also this application is 
currently in beta so please use responsible

-Thanks
Matthew W. Whited 

=================================================

You must have the includes for ncpfs installed on 
your system to compile this application.  Make sure
to run 'make install-dev' on ncpfs.

The command line for this application requires you 
to enter the mount point for your bordermanager server.
I recommend the SYS volume becasue it has several
public access directories.

Please note that the ClientTrust protocol was not
developed for multi-user systems.  Any user 
authenticated to you computer will be reported as 
the first person to authenticate.  This is a limitation
of the protocol and can not be fixed by me.

************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h> /* close() */
#include <string.h> /* memset() */
#include <stdlib.h>
#include <ncp/nwcalls.h> /* part of ncpfs */
#include <ncp/ncplib.h>  /* part of ncpfs */
#include <fcntl.h>
#include <syslog.h>

#define LOCAL_SERVER_PORT 3024
#define MAX_MSG 400
#define OUT_MSG 18


char *strmid(unsigned int start, char *stringin, unsigned int length) {
        int strmidx;
        char *stringout=(char *)malloc(length+1);

        for(strmidx = start; strmidx < length+start; strmidx++) {
              stringout[strmidx-start] = stringin[strmidx];
        }
	//stringout[strmidx-start+1]=(char)0;
        return stringout;
}

char *strrnul(char *stringin, unsigned int length) {
	int strrnulx;
	int strrnuly=0;

	char *stringout=(char *)malloc(length+1);

	for(strrnulx=0; strrnulx<=length; strrnulx+=2) {
	//		printf(" %X",stringin[strrnulx]);
			stringout[strrnuly]=stringin[strrnulx];
			strrnuly++;
	}
	//stringout[strrnuly]="\0";
	//printf("\n");
	return stringout;
}

void usage(int argc, char *argv[]) {
	printf("\n");
	printf("%s /mnt/bordermanager/SYS &" , argv[0]);
	printf("\n\n");

}

int main(int argc, char *argv[]) {

	//printf("Main arg count %d\n", argc);//DEBUG TC

	if (argc <= 1) {
		usage(argc, argv);
		exit (1);
	}

	/* ====== try to daemonize ====== */
	pid_t pid, sid;

        /* Fork off the parent process */
        /* TC pid = fork();
        if (pid < 0) {
                exit(EXIT_FAILURE);
        }*/
        /* If we got a good PID, then
           we can exit the parent process. */
        /* TC if (pid > 0) {
                exit(EXIT_SUCCESS);
        }*/

        /* Change the file mode mask */
        umask(0);       
        
        /* Open any logs here */
        
        /* Create a new SID for the child process */
        /* TC sid = setsid();
        if (sid < 0) {*/
                /* Log any failures here */
        /* TC         exit(EXIT_FAILURE);
        }*/

	/* ====== done daemonize ====== */

	int sd, rc, n, cliLen, servLen;
	struct sockaddr_in cliAddr, servAddr;
	char *msg=(char *)malloc(MAX_MSG+1);

  /* socket creation */
  	sd=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
  	if(sd<0) {
		printf("%s: cannot open socket \n",argv[0]);
		exit(1);
	}

  /* bind local server port */
  	servAddr.sin_family = AF_INET;
  	servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  	servAddr.sin_port = htons(LOCAL_SERVER_PORT);

  	rc = bind (sd, (struct sockaddr *) &servAddr,sizeof(servAddr));
  	if(rc<0) {
		printf("%s: cannot bind port number %d because %d \n",
			argv[0], LOCAL_SERVER_PORT, rc);
		exit(1);
	}

  /* server infinite loop */
        int x;
	int y;
	char *cthead=(char *)malloc(5);
        char *ctuid=(char *)malloc(5);
        char *ctbody=(char *)malloc(MAX_MSG-8);
	char *ctbodyu=(char *)malloc(MAX_MSG-8);

	char *ctheadreq=(char *)malloc(5);
	char *ctheadrep=(char *)malloc(5);
	char *ctheadna=(char *)malloc(5);

	char *msgout=(char *)malloc(OUT_MSG+1);

	ctheadreq[0]=17;
	ctheadreq[1]=17;
	ctheadreq[2]=17;
	ctheadreq[3]=17;

	ctheadrep[0]=34;
	ctheadrep[1]=34;
	ctheadrep[2]=34;
	ctheadrep[3]=34;

	ctheadna[0]=51;
	ctheadna[1]=51;
	ctheadna[2]=51;
	ctheadna[3]=51;

	servLen = sizeof(servAddr);

	NWCCODE ncperr;
	unsigned char ncpbuf[10];

	NWCONN_HANDLE ncpconn;

	NWCCODE ncperropen;

	char *connloc; // = "/root/.cl4others/BMSYS";
	connloc=argv[1];	

	ncpbuf[0]=0;
	ncpbuf[1]=8;
	ncpbuf[2]=0; /* this will be filled in by Client Trust proto */
	ncpbuf[3]=0;
	ncpbuf[4]=0;
	ncpbuf[5]=0;
	
	while(1) {

	

    /* init buffer */
		memset(msg,0x0,MAX_MSG);
       		memset(cthead,0x0,5);
        	memset(ctuid,0x0,5);
        	memset(ctbody,0x0,MAX_MSG);
		memset(msgout,0x0,OUT_MSG);

    /* receive message */
		cliLen = sizeof(cliAddr);

		n = recvfrom(sd, msg, MAX_MSG, 0,
			(struct sockaddr *) &cliAddr, (socklen_t *)&cliLen); //(i added socketlen_t *) to fix an error but im not sure this is the solution
		if(n<0) {
			printf("%s: cannot receive data \n",argv[0]);
			continue;
		}

    /* received message */
		/*		 
		for (x=0;x<56;x++) {
			printf(" %X", msg[x]);
		}
		printf("\n");
		*/
	
		ctbodyu = strmid( 16, msg, n-16);
		ctbody = strrnul(ctbodyu, n-16);

		ncpbuf[2]=msg[12];

		ctuid[0]=ncpbuf[6]=msg[4];
		ctuid[1]=ncpbuf[7]=msg[5];
		ctuid[2]=ncpbuf[8]=msg[6];
		ctuid[3]=ncpbuf[9]=msg[7];

		cthead[0]=msg[0];
		cthead[1]=msg[1];
		cthead[2]=msg[2];
		cthead[3]=msg[3];

		if (cthead[0]==ctheadreq[0] 
			&& cthead[1]==ctheadreq[1]
			&& cthead[2]==ctheadreq[2] 
			&& cthead[3]==ctheadreq[3]) 
		{

			//printf("%s\n",ctbody);
			
		/* prepare packet */
			msgout[4]=ctuid[0];
			msgout[5]=ctuid[1];
			msgout[6]=ctuid[2];
			msgout[7]=ctuid[3];

			for (x=8;x<(8+10+1);x++) {
				msgout[x]=0;
			}
		
		/* Find Connection to Proxy Server */
			ncperropen = ncp_open_mount(connloc, &ncpconn);

			if (ncperropen) { /* Connection Doesn't Exist */
				msgout[0]=ctheadna[0];
				msgout[1]=ctheadna[1];
				msgout[2]=ctheadna[2];
				msgout[3]=ctheadna[3];
				printf("Not Auth\n");

			} else { /* Connection Exists */
				msgout[0]=ctheadrep[0];
				msgout[1]=ctheadrep[1];
				msgout[2]=ctheadrep[2];
				msgout[3]=ctheadrep[3];
				//printf("Auth\n");

			}
		
		/* reply to Bordermanager with Client Trust auth packet */
			n = sendto(sd, msgout, 18, 0,
				(struct sockaddr *) &cliAddr, 
				sizeof(cliAddr));
			if(n == -1 ) {
				printf("%s could not send reply\n\n", argv[0]);
				exit (1);
			} else {
				if(!ncperropen) {

			/* send NCP extension to Bordermanager server after clienttrust auth */
					/*
					for(x=0;x<10;x++) {
						printf(" %X", ncpbuf[x]);
					}
					printf("\n");
					*/

					ncperr = NWRequestSimple(ncpconn, 37, ncpbuf, 
						sizeof(ncpbuf), NULL);
					if(ncperr) { 
						printf("Could Not Send NCPX Packet\n"); 
						printf("%X err code\n", ncperr);
						printf("%X conn\n", ncpconn);
						printf("%u ncpbuf - %d len\n",ncpbuf,sizeof(ncpbuf));
						printf("\n");
						if (ncperr == 5){//TC - restart but claims port in use because -1
						//  printf("attempting restart...\n");
						//  main(argc,argv);
							printf("Error code 5, exiting\n");
							exit(1);
						}
					}
				} else {
					printf("NCP Not Sent No Auth\n");
				}
			}
		}
		/*
		for (x=0;x<18;x++) {
			printf(" %X", msgout[x]);
		}
		printf("\n");
		*/
	}
	return 0;
}



